# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

try:
    import grpc
    from grpc._channel import (
        _UnaryUnaryMultiCallable,
        _StreamUnaryMultiCallable,
        _UnaryStreamMultiCallable,
        _StreamStreamMultiCallable,
    )

    import wrapt

    from instana.log import logger
    from instana.singletons import tracer
    from instana.propagators.format import Format
    from instana.span.span import get_current_span

    SUPPORTED_TYPES = [
        _UnaryUnaryMultiCallable,
        _StreamUnaryMultiCallable,
        _UnaryStreamMultiCallable,
        _StreamStreamMultiCallable,
    ]

    def collect_attributes(span, instance, argv, kwargs):
        try:
            span.set_attribute("rpc.flavor", "grpc")

            if type(instance) in SUPPORTED_TYPES:
                method = instance._method.decode()
                target = instance._channel.target().decode()
            elif type(argv[0]) is grpc._cython.cygrpc.RequestCallEvent:
                method = argv[0].call_details.method.decode()
                target = argv[0].call_details.host.decode()
            elif len(argv) > 2:
                method = argv[2][2][1]._method.decode()
                target = argv[2][2][1]._channel.target().decode()

            span.set_attribute("rpc.call", method)

            if ":///" in target:
                _, target, *_ = target.split(":///")
            parts = target.split(":")
            if len(parts) == 2:
                span.set_attribute("rpc.host", parts[0])
                span.set_attribute("rpc.port", parts[1])
        except Exception:
            logger.debug("grpc.collect_attributes non-fatal error", exc_info=True)
        return span

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryUnaryMultiCallable.with_call")
    def unary_unary_with_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "unary")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryUnaryMultiCallable.future")
    def unary_unary_future_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "unary")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryUnaryMultiCallable.__call__")
    def unary_unary_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context, record_exception=False
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "unary")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._channel", "_StreamUnaryMultiCallable.__call__")
    def stream_unary_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "stream")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper(
        "grpc._channel", "_StreamUnaryMultiCallable.with_call"
    )
    def stream_unary_with_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "stream")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._channel", "_StreamUnaryMultiCallable.future")
    def stream_unary_future_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "stream")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryStreamMultiCallable.__call__")
    def unary_stream_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "stream")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper(
        "grpc._channel", "_StreamStreamMultiCallable.__call__"
    )
    def stream_stream_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = get_current_span()

        # If we're not tracing, just return
        if not parent_span.is_recording():
            return wrapped(*argv, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "rpc-client", span_context=parent_context
        ) as span:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(
                    span.context,
                    Format.BINARY,
                    kwargs["metadata"],
                    disable_w3c_trace_context=True,
                )
                collect_attributes(span, instance, argv, kwargs)
                span.set_attribute("rpc.call_type", "stream")

                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper("grpc._server", "_call_behavior")
    def call_behavior_with_instana(wrapped, instance, argv, kwargs):
        # Prep any incoming context headers
        metadata = argv[0].invocation_metadata
        metadata_dict = {}
        for c in metadata:
            metadata_dict[c.key] = c.value

        ctx = tracer.extract(
            Format.BINARY, metadata_dict, disable_w3c_trace_context=True
        )

        with tracer.start_as_current_span("rpc-server", span_context=ctx) as span:
            try:
                collect_attributes(span, instance, argv, kwargs)
                rv = wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    logger.debug("Instrumenting grpcio")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import time
from typing import Generator
from unittest.mock import patch

import pytest

from instana.singletons import agent, tracer
from tests.helpers import testenv, get_first_span_by_name, get_first_span_by_filter

from couchbase.admin import Admin
from couchbase.cluster import Cluster
from couchbase.bucket import Bucket
from couchbase.exceptions import (
    CouchbaseTransientError,
    HTTPError,
    KeyExistsError,
    NotFoundError,
)
import couchbase.subdocument as SD
from couchbase.n1ql import N1QLQuery

# Delete any pre-existing buckets.  Create new.
cb_adm = Admin(
    testenv["couchdb_username"],
    testenv["couchdb_password"],
    host=testenv["couchdb_host"],
    port=8091,
)

# Make sure a test bucket exists
try:
    cb_adm.bucket_create("travel-sample")
    cb_adm.wait_ready("travel-sample", timeout=30)
except HTTPError:
    pass


class TestStandardCouchDB:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.cluster = Cluster("couchbase://%s" % testenv["couchdb_host"])
        self.bucket = Bucket(
            "couchbase://%s/travel-sample" % testenv["couchdb_host"],
            username=testenv["couchdb_username"],
            password=testenv["couchdb_password"],
        )
        self.bucket.upsert("test-key", 1)
        time.sleep(0.5)
        self.recorder.clear_spans()
        yield
        agent.options.allow_exit_as_root = False

    def test_vanilla_get(self) -> None:
        res = self.bucket.get("test-key")
        assert res

    def test_upsert(self) -> None:
        res = None
        with tracer.start_as_current_span("test"):
            res = self.bucket.upsert("test_upsert", 1)

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "upsert"

    def test_upsert_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        res = self.bucket.upsert("test_upsert", 1)

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        assert not cb_span.p

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "upsert"

    def test_upsert_multi(self) -> None:
        res = None

        kvs = {}
        kvs["first_test_upsert_multi"] = 1
        kvs["second_test_upsert_multi"] = 1

        with tracer.start_as_current_span("test"):
            res = self.bucket.upsert_multi(kvs)

        assert res
        assert res["first_test_upsert_multi"].success
        assert res["second_test_upsert_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "upsert_multi"

    def test_insert_new(self) -> None:
        res = None
        try:
            self.bucket.remove("test_insert_new")
        except NotFoundError:
            pass

        with tracer.start_as_current_span("test"):
            res = self.bucket.insert("test_insert_new", 1)

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "insert"

    def test_insert_existing(self) -> None:
        res = None
        try:
            self.bucket.insert("test_insert", 1)
        except KeyExistsError:
            pass

        try:
            with tracer.start_as_current_span("test"):
                res = self.bucket.insert("test_insert", 1)
        except KeyExistsError:
            pass

        assert not res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert cb_span.ec == 1
        # Just search for the substring of the exception class
        found = cb_span.data["couchbase"]["error"].find("_KeyExistsError")
        assert not found == -1

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "insert"

    def test_insert_multi(self) -> None:
        res = None

        kvs = {}
        kvs["first_test_upsert_multi"] = 1
        kvs["second_test_upsert_multi"] = 1

        try:
            self.bucket.remove("first_test_upsert_multi")
            self.bucket.remove("second_test_upsert_multi")
        except NotFoundError:
            pass

        with tracer.start_as_current_span("test"):
            res = self.bucket.insert_multi(kvs)

        assert res
        assert res["first_test_upsert_multi"].success
        assert res["second_test_upsert_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "insert_multi"

    def test_replace(self) -> None:
        res = None
        try:
            self.bucket.insert("test_replace", 1)
        except KeyExistsError:
            pass

        with tracer.start_as_current_span("test"):
            res = self.bucket.replace("test_replace", 2)

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "replace"

    def test_replace_non_existent(self) -> None:
        res = None

        try:
            self.bucket.remove("test_replace")
        except NotFoundError:
            pass

        try:
            with tracer.start_as_current_span("test"):
                res = self.bucket.replace("test_replace", 2)
        except NotFoundError:
            pass

        assert not res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert cb_span.ec == 1
        # Just search for the substring of the exception class
        found = cb_span.data["couchbase"]["error"].find("NotFoundError")
        assert not found == -1

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "replace"

    def test_replace_multi(self) -> None:
        res = None

        kvs = {}
        kvs["first_test_replace_multi"] = 1
        kvs["second_test_replace_multi"] = 1

        self.bucket.upsert("first_test_replace_multi", "one")
        self.bucket.upsert("second_test_replace_multi", "two")

        with tracer.start_as_current_span("test"):
            res = self.bucket.replace_multi(kvs)

        assert res
        assert res["first_test_replace_multi"].success
        assert res["second_test_replace_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "replace_multi"

    def test_append(self) -> None:
        self.bucket.upsert("test_append", "one")

        res = None
        with tracer.start_as_current_span("test"):
            res = self.bucket.append("test_append", "two")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "append"

    def test_append_multi(self) -> None:
        res = None

        kvs = dict()
        kvs["first_test_append_multi"] = "ok1"
        kvs["second_test_append_multi"] = "ok2"

        self.bucket.upsert("first_test_append_multi", "one")
        self.bucket.upsert("second_test_append_multi", "two")

        with tracer.start_as_current_span("test"):
            res = self.bucket.append_multi(kvs)

        assert res
        assert res["first_test_append_multi"].success
        assert res["second_test_append_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "append_multi"

    def test_prepend(self) -> None:
        self.bucket.upsert("test_prepend", "one")

        res = None
        with tracer.start_as_current_span("test"):
            res = self.bucket.prepend("test_prepend", "two")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "prepend"

    def test_prepend_multi(self) -> None:
        res = None

        kvs = {}
        kvs["first_test_prepend_multi"] = "ok1"
        kvs["second_test_prepend_multi"] = "ok2"

        self.bucket.upsert("first_test_prepend_multi", "one")
        self.bucket.upsert("second_test_prepend_multi", "two")

        with tracer.start_as_current_span("test"):
            res = self.bucket.prepend_multi(kvs)

        assert res
        assert res["first_test_prepend_multi"].success
        assert res["second_test_prepend_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "prepend_multi"

    def test_get(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.get("test-key")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "get"

    def test_rget(self) -> None:
        res = None

        try:
            with tracer.start_as_current_span("test"):
                res = self.bucket.rget("test-key", replica_index=None)
        except CouchbaseTransientError:
            pass

        assert not res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert cb_span.ec == 1
        # Just search for the substring of the exception class
        found = cb_span.data["couchbase"]["error"].find("CouchbaseTransientError")
        assert found != -1

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "rget"

    def test_get_not_found(self) -> None:
        res = None
        try:
            self.bucket.remove("test_get_not_found")
        except NotFoundError:
            pass

        try:
            with tracer.start_as_current_span("test"):
                res = self.bucket.get("test_get_not_found")
        except NotFoundError:
            pass

        assert not res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert cb_span.ec == 1
        # Just search for the substring of the exception class
        found = cb_span.data["couchbase"]["error"].find("NotFoundError")
        assert found != -1

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "get"

    def test_get_multi(self) -> None:
        res = None

        self.bucket.upsert("first_test_get_multi", "one")
        self.bucket.upsert("second_test_get_multi", "two")

        with tracer.start_as_current_span("test"):
            res = self.bucket.get_multi(
                ["first_test_get_multi", "second_test_get_multi"]
            )

        assert res
        assert res["first_test_get_multi"].success
        assert res["second_test_get_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "get_multi"

    def test_touch(self) -> None:
        res = None
        self.bucket.upsert("test_touch", 1)

        with tracer.start_as_current_span("test"):
            res = self.bucket.touch("test_touch")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "touch"

    def test_touch_multi(self) -> None:
        res = None

        self.bucket.upsert("first_test_touch_multi", "one")
        self.bucket.upsert("second_test_touch_multi", "two")

        with tracer.start_as_current_span("test"):
            res = self.bucket.touch_multi(
                ["first_test_touch_multi", "second_test_touch_multi"]
            )

        assert res
        assert res["first_test_touch_multi"].success
        assert res["second_test_touch_multi"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "touch_multi"

    def test_lock(self) -> None:
        res = None
        self.bucket.upsert("test_lock_unlock", "lock_this")

        with tracer.start_as_current_span("test"):
            rv = self.bucket.lock("test_lock_unlock", ttl=5)
            assert rv
            assert rv.success

            # upsert automatically unlocks the key
            res = self.bucket.upsert("test_lock_unlock", "updated", rv.cas)
            assert res
            assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        def filter(span):
            return span.n == "couchbase" and span.data["couchbase"]["type"] == "lock"

        cb_lock_span = get_first_span_by_filter(spans, filter)
        assert cb_lock_span

        def filter(span):
            return span.n == "couchbase" and span.data["couchbase"]["type"] == "upsert"

        cb_upsert_span = get_first_span_by_filter(spans, filter)
        assert cb_upsert_span

        # Same traceId and parent relationship
        assert cb_lock_span.t == test_span.t
        assert cb_upsert_span.t == test_span.t

        assert cb_lock_span.p == test_span.s
        assert cb_upsert_span.p == test_span.s

        assert cb_lock_span.stack
        assert not cb_lock_span.ec
        assert cb_upsert_span.stack
        assert not cb_upsert_span.ec

        assert (
            cb_lock_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_lock_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_lock_span.data["couchbase"]["type"] == "lock"
        assert (
            cb_upsert_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_upsert_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_upsert_span.data["couchbase"]["type"] == "upsert"

    def test_lock_unlock(self) -> None:
        res = None
        self.bucket.upsert("test_lock_unlock", "lock_this")

        with tracer.start_as_current_span("test"):
            rv = self.bucket.lock("test_lock_unlock", ttl=5)
            assert rv
            assert rv.success

            # upsert automatically unlocks the key
            res = self.bucket.unlock("test_lock_unlock", rv.cas)
            assert res
            assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        def filter(span):
            return span.n == "couchbase" and span.data["couchbase"]["type"] == "lock"

        cb_lock_span = get_first_span_by_filter(spans, filter)
        assert cb_lock_span

        def filter(span):
            return span.n == "couchbase" and span.data["couchbase"]["type"] == "unlock"

        cb_unlock_span = get_first_span_by_filter(spans, filter)
        assert cb_unlock_span

        # Same traceId and parent relationship
        assert cb_lock_span.t == test_span.t
        assert cb_unlock_span.t == test_span.t

        assert cb_lock_span.p == test_span.s
        assert cb_unlock_span.p == test_span.s

        assert cb_lock_span.stack
        assert not cb_lock_span.ec
        assert cb_unlock_span.stack
        assert not cb_unlock_span.ec

        assert (
            cb_lock_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_lock_span.data["couchbase"]["bucket"], "travel-sample"
        assert cb_lock_span.data["couchbase"]["type"], "lock"
        assert (
            cb_unlock_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_unlock_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_unlock_span.data["couchbase"]["type"] == "unlock"

    def test_lock_unlock_muilti(self) -> None:
        res = None
        self.bucket.upsert("test_lock_unlock_multi_1", "lock_this")
        self.bucket.upsert("test_lock_unlock_multi_2", "lock_this")

        keys_to_lock = ("test_lock_unlock_multi_1", "test_lock_unlock_multi_2")

        with tracer.start_as_current_span("test"):
            rv = self.bucket.lock_multi(keys_to_lock, ttl=5)
            assert rv
            assert rv["test_lock_unlock_multi_1"].success
            assert rv["test_lock_unlock_multi_2"].success

            res = self.bucket.unlock_multi(rv)
            assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        def filter(span):
            return (
                span.n == "couchbase" and span.data["couchbase"]["type"] == "lock_multi"
            )

        cb_lock_span = get_first_span_by_filter(spans, filter)
        assert cb_lock_span

        def filter(span):
            return (
                span.n == "couchbase"
                and span.data["couchbase"]["type"] == "unlock_multi"
            )

        cb_unlock_span = get_first_span_by_filter(spans, filter)
        assert cb_unlock_span

        # Same traceId and parent relationship
        assert cb_lock_span.t == test_span.t
        assert cb_unlock_span.t == test_span.t

        assert cb_lock_span.p == test_span.s
        assert cb_unlock_span.p == test_span.s

        assert cb_lock_span.stack
        assert not cb_lock_span.ec
        assert cb_unlock_span.stack
        assert not cb_unlock_span.ec

        assert (
            cb_lock_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_lock_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_lock_span.data["couchbase"]["type"] == "lock_multi"
        assert (
            cb_unlock_span.data["couchbase"]["hostname"]
            == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_unlock_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_unlock_span.data["couchbase"]["type"] == "unlock_multi"

    def test_remove(self) -> None:
        res = None
        self.bucket.upsert("test_remove", 1)

        with tracer.start_as_current_span("test"):
            res = self.bucket.remove("test_remove")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "remove"

    def test_remove_multi(self) -> None:
        res = None
        self.bucket.upsert("test_remove_multi_1", 1)
        self.bucket.upsert("test_remove_multi_2", 1)

        keys_to_remove = ("test_remove_multi_1", "test_remove_multi_2")

        with tracer.start_as_current_span("test"):
            res = self.bucket.remove_multi(keys_to_remove)

        assert res
        assert res["test_remove_multi_1"].success
        assert res["test_remove_multi_2"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "remove_multi"

    def test_counter(self) -> None:
        res = None
        self.bucket.upsert("test_counter", 1)

        with tracer.start_as_current_span("test"):
            res = self.bucket.counter("test_counter", delta=10)

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "counter"

    def test_counter_multi(self) -> None:
        res = None
        self.bucket.upsert("first_test_counter", 1)
        self.bucket.upsert("second_test_counter", 1)

        with tracer.start_as_current_span("test"):
            res = self.bucket.counter_multi(
                ("first_test_counter", "second_test_counter")
            )

        assert res
        assert res["first_test_counter"].success
        assert res["second_test_counter"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "counter_multi"

    def test_mutate_in(self) -> None:
        res = None
        self.bucket.upsert(
            "king_arthur",
            {
                "name": "Arthur",
                "email": "kingarthur@couchbase.com",
                "interests": ["Holy Grail", "African Swallows"],
            },
        )

        with tracer.start_as_current_span("test"):
            res = self.bucket.mutate_in(
                "king_arthur",
                SD.array_addunique("interests", "Cats"),
                SD.counter("updates", 1),
            )

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "mutate_in"

    def test_lookup_in(self) -> None:
        res = None
        self.bucket.upsert(
            "king_arthur",
            {
                "name": "Arthur",
                "email": "kingarthur@couchbase.com",
                "interests": ["Holy Grail", "African Swallows"],
            },
        )

        with tracer.start_as_current_span("test"):
            res = self.bucket.lookup_in(
                "king_arthur", SD.get("email"), SD.get("interests")
            )

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "lookup_in"

    def test_stats(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.stats()

        assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "stats"

    def test_ping(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.ping()

        assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "ping"

    def test_diagnostics(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.diagnostics()

        assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "diagnostics"

    def test_observe(self) -> None:
        res = None
        self.bucket.upsert("test_observe", 1)

        with tracer.start_as_current_span("test"):
            res = self.bucket.observe("test_observe")

        assert res
        assert res.success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "observe"

    def test_observe_multi(self) -> None:
        res = None
        self.bucket.upsert("test_observe_multi_1", 1)
        self.bucket.upsert("test_observe_multi_2", 1)

        keys_to_observe = ("test_observe_multi_1", "test_observe_multi_2")

        with tracer.start_as_current_span("test"):
            res = self.bucket.observe_multi(keys_to_observe)

        assert res
        assert res["test_observe_multi_1"].success
        assert res["test_observe_multi_2"].success

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "observe_multi"

    def test_query_with_instana_tracing_off(self) -> None:
        res = None

        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.couchbase_inst.tracing_is_off", return_value=True
        ):
            res = self.bucket.n1ql_query("SELECT 1")
        assert res

    def test_query_with_instana_exception(self) -> None:
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.couchbase_inst.collect_attributes",
            side_effect=Exception("test-error"),
        ):
            self.bucket.n1ql_query("SELECT 1")

        spans = self.recorder.queued_spans()
        cb_span = get_first_span_by_name(spans, "couchbase")

        assert cb_span.data["couchbase"]["error"] == "Exception('test-error')"

    def test_raw_n1ql_query(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.n1ql_query("SELECT 1")

        assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "n1ql_query"
        assert cb_span.data["couchbase"]["sql"] == "SELECT 1"

    def test_n1ql_query(self) -> None:
        res = None

        with tracer.start_as_current_span("test"):
            res = self.bucket.n1ql_query(
                N1QLQuery(
                    'SELECT name FROM `travel-sample` WHERE brewery_id ="mishawaka_brewing"'
                )
            )

        assert res

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cb_span = get_first_span_by_name(spans, "couchbase")
        assert cb_span

        # Same traceId and parent relationship
        assert cb_span.t == test_span.t
        assert cb_span.p == test_span.s

        assert cb_span.stack
        assert not cb_span.ec

        assert (
            cb_span.data["couchbase"]["hostname"] == f"{testenv['couchdb_host']}:8091"
        )
        assert cb_span.data["couchbase"]["bucket"] == "travel-sample"
        assert cb_span.data["couchbase"]["type"] == "n1ql_query"
        assert (
            cb_span.data["couchbase"]["sql"]
            == 'SELECT name FROM `travel-sample` WHERE brewery_id ="mishawaka_brewing"'
        )



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import time
from contextlib import contextmanager
from typing import Any, Callable, Dict, Iterator, Tuple

import wrapt
from opentelemetry.trace import use_span
from opentelemetry.trace.status import StatusCode

from instana.configurator import config
from instana.log import logger
from instana.span.span import InstanaSpan
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    import asyncio

    @wrapt.patch_function_wrapper("asyncio", "ensure_future")
    def ensure_future_with_instana(
        wrapped: Callable[..., asyncio.ensure_future],
        instance: object,
        argv: Tuple[object, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if (
            not config["asyncio_task_context_propagation"]["enabled"]
            or tracing_is_off()
        ):
            return wrapped(*argv, **kwargs)

        with _start_as_current_async_span() as span:
            try:
                span.set_status(StatusCode.OK)
                return wrapped(*argv, **kwargs)
            except Exception as exc:
                logger.debug(f"asyncio ensure_future_with_instana error: {exc}")

    if hasattr(asyncio, "create_task"):

        @wrapt.patch_function_wrapper("asyncio", "create_task")
        def create_task_with_instana(
            wrapped: Callable[..., asyncio.create_task],
            instance: object,
            argv: Tuple[object, Tuple[object, ...]],
            kwargs: Dict[str, Any],
        ) -> object:
            if (
                not config["asyncio_task_context_propagation"]["enabled"]
                or tracing_is_off()
            ):
                return wrapped(*argv, **kwargs)

            with _start_as_current_async_span() as span:
                try:
                    span.set_status(StatusCode.OK)
                    return wrapped(*argv, **kwargs)
                except Exception as exc:
                    logger.debug(f"asyncio create_task_with_instana error: {exc}")

    @contextmanager
    def _start_as_current_async_span() -> Iterator[InstanaSpan]:
        """
        Creates and yield a special InstanaSpan to only propagate the Asyncio
        context.
        """
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        _time = time.time_ns()

        span = InstanaSpan(
            name="asyncio",
            context=parent_context,
            span_processor=tracer.span_processor,
            start_time=_time,
            end_time=_time,
        )
        with use_span(
            span,
            end_on_exit=False,
            record_exception=False,
            set_status_on_exception=False,
        ) as span:
            yield span

    logger.debug("Instrumenting asyncio")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import logging
from typing import Generator
from unittest.mock import patch
import pytest
import redis

from instana.span.span import get_current_span
from tests.helpers import testenv
from instana.singletons import agent, tracer


class TestRedis:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.client = redis.Redis(host=testenv["redis_host"], db=testenv["redis_db"])
        yield
        agent.options.allow_exit_as_root = False

    def test_set_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.set("fooy", "barY")
            result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_set_get_as_root_span(self) -> None:
        agent.options.allow_exit_as_root = True

        self.client.set("foox", "barX")
        self.client.set("fooy", "barY")
        result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Parent relationships
        assert not rs1_span.p
        assert not rs2_span.p
        assert not rs3_span.p

        # Error logging
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_set_incr_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("counter", "10")
            self.client.incr("counter")
            result = self.client.get("counter")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"11"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "INCRBY"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_old_redis_client(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.set("fooy", "barY")
            result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )

        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_pipelined_requests(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            result = pipe.execute()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        assert result == [True, True, b"barX"]

        rs1_span = spans[0]
        test_span = spans[1]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "PIPELINE"
        assert rs1_span.data["redis"]["subCommands"] == ["SET", "SET", "GET"]
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

    @patch(
        "instana.instrumentation.redis.collect_attributes",
        side_effect=Exception("test-error"),
    )
    @patch("instana.span.span.InstanaSpan.record_exception")
    def test_execute_command_with_instana_exception(self, mock_record_func, _) -> None:
        with tracer.start_as_current_span("test"), pytest.raises(
            Exception, match="test-error"
        ):
            self.client.set("counter", "10")
        mock_record_func.assert_called()

    def test_execute_comand_with_instana_tracing_off(self) -> None:
        with tracer.start_as_current_span("redis"):
            response = self.client.set("counter", "10")
            assert response

    def test_execute_with_instana_tracing_off(self) -> None:
        result = None
        with tracer.start_as_current_span("redis"):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            result = pipe.execute()
            assert result == [True, True, b"barX"]

    def test_execute_with_instana_exception(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.redis.collect_attributes",
            side_effect=Exception("test-error"),
        ):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            pipe.execute()
        assert "Error collecting pipeline commands" in caplog.messages



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for Starlette
https://www.starlette.io/
"""

from typing import Any, Callable, Dict, Tuple

try:
    import starlette
    import wrapt
    from starlette.middleware import Middleware
    import starlette.applications

    from instana.instrumentation.asgi import InstanaASGIMiddleware
    from instana.log import logger

    @wrapt.patch_function_wrapper("starlette.applications", "Starlette.__init__")
    def init_with_instana(
        wrapped: Callable[..., starlette.applications.Starlette.__init__],
        instance: starlette.applications.Starlette,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> None:
        middleware = kwargs.get("middleware")
        if middleware is None:
            kwargs["middleware"] = [Middleware(InstanaASGIMiddleware)]
        elif isinstance(middleware, list):
            middleware.append(Middleware(InstanaASGIMiddleware))

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting Starlette")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2017


from instana.instrumentation.asgi import InstanaASGIMiddleware  # noqa: F401
from instana.instrumentation.wsgi import InstanaWSGIMiddleware  # noqa: F401



# (c) Copyright IBM Corp. 2024

from typing import Generator
from unittest.mock import patch

import pytest

from instana.agent.host import HostAgent
from instana.collector.helpers.runtime import RuntimeHelper
from instana.collector.host import HostCollector


class TestRuntimeHelper:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.helper = RuntimeHelper(
            collector=HostCollector(
                HostAgent(),
            ),
        )
        yield
        self.helper = None

    def test_default_while_gc_disabled(self) -> None:
        import gc

        gc.disable()
        helper = RuntimeHelper(collector=HostCollector(HostAgent()))
        assert helper.previous_gc_count is None

    def test_collect_metrics(self) -> None:
        response = self.helper.collect_metrics()
        assert response[0]["name"] == "com.instana.plugin.python"

    def test_collect_runtime_snapshot_default(self) -> None:
        plugin_data = self.helper.collect_metrics()
        self.helper._collect_runtime_snapshot(plugin_data[0])
        assert plugin_data[0]["name"] == "com.instana.plugin.python"
        assert plugin_data[0]["data"]["snapshot"]["m"] == "Manual"
        assert len(plugin_data[0]["data"]) == 3

    def test_collect_runtime_snapshot_autowrapt(self) -> None:
        with patch(
            "instana.collector.helpers.runtime.is_autowrapt_instrumented",
            return_value=True,
        ):
            plugin_data = self.helper.collect_metrics()
            self.helper._collect_runtime_snapshot(plugin_data[0])
            assert plugin_data[0]["name"] == "com.instana.plugin.python"
            assert plugin_data[0]["data"]["snapshot"]["m"] == "Autowrapt"
            assert len(plugin_data[0]["data"]) == 3

    def test_collect_runtime_snapshot_webhook(self) -> None:
        with patch(
            "instana.collector.helpers.runtime.is_webhook_instrumented",
            return_value=True,
        ):
            plugin_data = self.helper.collect_metrics()
            self.helper._collect_runtime_snapshot(plugin_data[0])
            assert plugin_data[0]["name"] == "com.instana.plugin.python"
            assert plugin_data[0]["data"]["snapshot"]["m"] == "AutoTrace"
            assert len(plugin_data[0]["data"]) == 3

    def test_collect_gc_metrics(self) -> None:
        plugin_data = self.helper.collect_metrics()

        self.helper._collect_gc_metrics(plugin_data[0], True)
        assert len(self.helper.previous["data"]["metrics"]["gc"]) == 6



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

class _TraceContextMixin:
    def assertTraceContextPropagated(self, parent_span, child_span):
        assert parent_span.t == child_span.t
        assert parent_span.s == child_span.p
        assert parent_span.s != child_span.s

    def assertErrorLogging(self, spans):
        for span in spans:
            assert not span.ec



# (c) Copyright IBM Corp. 2024

import os
import sys
from tests.apps.aiohttp_app2.app import aiohttp_server as server
from tests.apps.utils import launch_background_thread

APP_THREAD = None

if not any((os.environ.get('GEVENT_STARLETTE_TEST'),
            os.environ.get('CASSANDRA_TEST'),
            sys.version_info < (3, 5, 3))):
    APP_THREAD = launch_background_thread(server, "AIOHTTP")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading
import time
from typing import Callable, Tuple

from instana.log import logger


class TimerWraper(object):
    def __init__(self) -> None:
        self.timer = None
        self.cancel_lock = threading.Lock()
        self.canceled = False

    def cancel(self) -> None:
        with self.cancel_lock:
            self.canceled = True
            self.timer.cancel()


def delay(
    timeout: float, func: Callable[..., object], *args: Tuple[object]
) -> threading.Timer:
    def func_wrapper() -> None:
        try:
            func(*args)
        except Exception:
            logger.error("Error in delayed function", exc_info=True)

    t = threading.Timer(timeout, func_wrapper, ())
    t.start()

    return t


def schedule(
    timeout: float, interval: float, func: Callable[..., object], *args: Tuple[object]
) -> TimerWraper:
    tw = TimerWraper()

    def func_wrapper() -> None:
        start = time.time()

        try:
            func(*args)
        except Exception:
            logger.error("Error in scheduled function", exc_info=True)

        with tw.cancel_lock:
            if not tw.canceled:
                tw.timer = threading.Timer(
                    abs(interval - (time.time() - start)), func_wrapper, ()
                )
                tw.timer.start()

    tw.timer = threading.Timer(timeout, func_wrapper, ())
    tw.timer.start()

    return tw



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

import os
import sys
import time

try:
    from django.urls import re_path, include
except ImportError:
    from django.conf.urls import url as re_path

from django.http import HttpResponse, Http404
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import SpanKind

from instana.singletons import tracer

filepath, extension = os.path.splitext(__file__)
os.environ["DJANGO_SETTINGS_MODULE"] = os.path.basename(filepath)
sys.path.insert(0, os.path.dirname(os.path.abspath(filepath)))

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = "^(myu#*^5v-9o$i-%6vnlwvy^#7&hspj$m3lcq#b$@__@+zd@c"
DEBUG = True
ALLOWED_HOSTS = ["testserver", "localhost"]
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]
ROOT_URLCONF = "app_django"
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
WSGI_APPLICATION = "app_django.wsgi.application"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}
AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = "/static/"


def index(request):
    return HttpResponse("Stan wuz here!")


def cause_error(request):
    raise Exception("This is a fake error: /cause-error")


def induce_exception(request):
    raise Exception("This is a fake error: /induce-exception")


def another(request):
    return HttpResponse("Stan wuz here!")


def not_found(request):
    raise Http404("Nothing here")


def complex(request):
    with tracer.start_as_current_span("asteroid") as pspan:
        pspan.set_attribute("component", "Python simple example app")
        pspan.set_attribute("span.kind", SpanKind.CLIENT)
        pspan.set_attribute("peer.hostname", "localhost")
        pspan.set_attribute(SpanAttributes.HTTP_URL, "/python/simple/one")
        pspan.set_attribute(SpanAttributes.HTTP_METHOD, "GET")
        pspan.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 200)
        pspan.add_event(name="complex_request", attributes={"foo": "bar"})
        time.sleep(0.2)

        with tracer.start_as_current_span("spacedust") as cspan:
            cspan.set_attribute("span.kind", SpanKind.CLIENT)
            cspan.set_attribute("peer.hostname", "localhost")
            cspan.set_attribute(SpanAttributes.HTTP_URL, "/python/simple/two")
            cspan.set_attribute(SpanAttributes.HTTP_METHOD, "POST")
            cspan.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 204)
            time.sleep(0.1)

    return HttpResponse("Stan wuz here!")


def response_with_headers(request):
    headers = {"X-Capture-This-Too": "this too", "X-Capture-That-Too": "that too"}
    return HttpResponse("Stan wuz here with headers!", headers=headers)


extra_patterns = [
    re_path(r"^induce_exception$", induce_exception, name="induce_exception"),
]

urlpatterns = [
    re_path(r"^$", index, name="index"),
    re_path(r"^cause_error$", cause_error, name="cause_error"),
    re_path(r"^another$", another),
    re_path(r"^not_found$", not_found, name="not_found"),
    re_path(
        r"^response_with_headers$", response_with_headers, name="response_with_headers"
    ),
    re_path(r"^exception$", include(extra_patterns)),
    re_path(r"^complex$", complex, name="complex"),
]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import time
import random
from typing import Union

from opentelemetry.trace.span import _SPAN_ID_MAX_VALUE, INVALID_SPAN_ID, INVALID_TRACE_ID

_rnd = random.Random()
_current_pid = 0


def generate_id() -> int:
    """Get a new ID.

    Returns:
        A 64-bit int for use as a Span or Trace ID.
    """
    global _current_pid

    pid = os.getpid()
    if _current_pid != pid:
        _current_pid = pid
        _rnd.seed(int(1000000 * time.time()) ^ pid)
    new_id = _rnd.randint(0, _SPAN_ID_MAX_VALUE)

    return new_id


def header_to_long_id(header: Union[bytes, str]) -> int:
    """
    We can receive headers in the following formats:
      1. unsigned base 16 hex string (or bytes) of variable length
      2. [eventual]

    :param header: the header to analyze, validate and convert (if needed)
    :return: a valid ID to be used internal to the tracer
    """
    if isinstance(header, bytes):
        header = header.decode('utf-8')

    if not isinstance(header, str):
        return INVALID_TRACE_ID

    if header.isdecimal():
        return header

    try:
        if len(header) < 16:
            # Left pad ID with zeros
            header = header.zfill(16)

        return int(header, 16)
    except ValueError:
        return INVALID_TRACE_ID


def header_to_id(header: Union[bytes, str]) -> int:
    """
    We can receive headers in the following formats:
      1. unsigned base 16 hex string (or bytes) of variable length
      2. [eventual]

    :param header: the header to analyze, validate and convert (if needed)
    :return: a valid ID to be used internal to the tracer
    """
    if isinstance(header, bytes):
        header = header.decode('utf-8')

    if not isinstance(header, str):
        return INVALID_SPAN_ID

    if header.isdecimal():
        return header

    try:
        length = len(header)
        if length < 16:
            # Left pad ID with zeros
            header = header.zfill(16)
        elif length > 16:
            # Phase 0: Discard everything but the last 16byte
            header = header[-16:]

        return int(header, 16)
    except ValueError:
        return INVALID_SPAN_ID


def hex_id(id: Union[int, str]) -> str:
    """
    Returns the hexadecimal representation of the given ID.
    Left pad with zeros when the length is not equal to 16
    """
    try:
        hex_id = hex(int(id))[2:]
        length = len(hex_id)
        # Left pad ID with zeros
        if length < 16:
            hex_id = hex_id.zfill(16)
        elif length > 16 and length < 32:
            hex_id = hex_id.zfill(32)
        return hex_id
    except ValueError: # Handles ValueError: invalid literal for int() with base 10:
        return id

def hex_id_limited(id: Union[int, str]) -> str:
    """
    Returns the hexadecimal representation of the given ID.
    Limit longer IDs to 16 characters 
    """
    try:
        hex_id = hex(int(id))[2:]
        length = len(hex_id)
        if length < 16:
            # Left pad ID with zeros
            hex_id = hex_id.zfill(16)
        elif length > 16:
            # Phase 0: Discard everything but the last 16byte
            hex_id = hex_id[-16:]
        return hex_id
    except ValueError: # Handles ValueError: invalid literal for int() with base 10:
        return id

def define_server_timing(trace_id: Union[int, str]) -> str:
    # Note: The key `intid` is short for Instana Trace ID.
    return f"intid;desc={hex_id_limited(trace_id)}"


def internal_id(id: Union[int, str]) -> int:
    """
    Returns a valid id to be used internally. Handles both str and int types.
    """
    if isinstance(id, int):
        return id
    
    length = len(id)

    if isinstance(id, str) and id.isdigit():
        if length == 16:
            return int(id, 16)
        else:
            return int(id)

    try:
        if length < 16:
            # Left pad ID with zeros
            id = id.zfill(16)

        # hex string -> int
        return int(id, 16)
    except ValueError:
        return INVALID_TRACE_ID
    
def internal_id_limited(id: Union[int, str]) -> int:
    """
    Returns a valid id to be used internally. Handles both str and int types.
    Note: Limits the hex string to 16 chars before conversion.
    """
    if isinstance(id, int):
        return id

    length = len(id)

    if isinstance(id, str) and id.isdigit():
        if length == 16:
            return int(id, 16)
        else:
            return int(id)

    try:
        if length < 16:
            # Left pad ID with zeros
            id = id.zfill(16)
        elif length > 16:
            # Phase 0: Discard everything but the last 16byte
            id = id[-16:]

        # hex string -> int
        return int(id, 16)
    except ValueError:
        return INVALID_SPAN_ID



# (c) Copyright IBM Corp. 2024

import time

from instana.span.readable_span import Event


def test_span_event_defaults():
    event_name = "test-span-event"
    event = Event(event_name)

    assert event
    assert isinstance(event, Event)
    assert event.name == event_name
    assert not event.attributes
    assert isinstance(event.timestamp, int)
    assert event.timestamp < time.time_ns()


def test_span_event():
    event_name = "test-span-event"
    attributes = {
        "field1": 1,
        "field2": "two",
    }
    timestamp = time.time_ns()

    event = Event(event_name, attributes, timestamp)

    assert event
    assert isinstance(event, Event)
    assert event.name == event_name
    assert event.attributes
    assert len(event.attributes) == 2
    assert "field1" in event.attributes.keys()
    assert "two" == event.attributes.get("field2")
    assert event.timestamp == timestamp


def test_event_with_params() -> None:
    name = "sample-event"
    attributes = ["attribute"]
    timestamp = time.time_ns()
    event = Event(name, attributes, timestamp)

    assert event.name == name
    assert event.attributes == attributes
    assert event.timestamp == timestamp



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import re
import flask
import wrapt
from typing import Callable, Tuple, Dict, Type, Union

from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry import context, trace

from instana.log import logger
from instana.singletons import agent, tracer
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import extract_custom_headers
from instana.propagators.format import Format

path_tpl_re = re.compile('<.*>')


def before_request_with_instana() -> None:
    try:
        env = flask.request.environ
        span_context = tracer.extract(Format.HTTP_HEADERS, env)

        span = tracer.start_span("wsgi", span_context=span_context)
        flask.g.span = span

        ctx = trace.set_span_in_context(span)
        token = context.attach(ctx)
        flask.g.token = token

        extract_custom_headers(span, env, format=True)

        span.set_attribute(SpanAttributes.HTTP_METHOD, flask.request.method)
        if "PATH_INFO" in env:
            span.set_attribute(SpanAttributes.HTTP_URL, env["PATH_INFO"])
        if "QUERY_STRING" in env and len(env["QUERY_STRING"]):
            scrubbed_params = strip_secrets_from_query(
                env["QUERY_STRING"],
                agent.options.secrets_matcher,
                agent.options.secrets_list,
            )
            span.set_attribute("http.params", scrubbed_params)
        if "HTTP_HOST" in env:
            span.set_attribute("http.host", env["HTTP_HOST"])

        if hasattr(flask.request.url_rule, "rule") and path_tpl_re.search(
            flask.request.url_rule.rule
        ):
            path_tpl = flask.request.url_rule.rule.replace("<", "{")
            path_tpl = path_tpl.replace(">", "}")
            span.set_attribute("http.path_tpl", path_tpl)
    except:
        logger.debug("Flask before_request", exc_info=True)

    return None


def after_request_with_instana(
    response: flask.wrappers.Response,
) -> flask.wrappers.Response:
    span = None
    try:
        # If we're not tracing, just return
        if not hasattr(flask.g, "span"):
            return response

        span = flask.g.span
        if span:

            if 500 <= response.status_code:
                span.mark_as_errored()

            span.set_attribute(
                SpanAttributes.HTTP_STATUS_CODE, int(response.status_code)
            )
            extract_custom_headers(span, response.headers, format=False)

            tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)
    except:
        logger.debug("Flask after_request", exc_info=True)
    finally:
        if span and span.is_recording():
            span.end()
            flask.g.span = None
    return response


def teardown_request_with_instana(*argv: Union[Exception, Type[Exception]]) -> None:
    """
    In the case of exceptions, after_request_with_instana isn't called
    so we capture those cases here.
    """
    if hasattr(flask.g, "span") and flask.g.span:
        if len(argv) > 0 and argv[0]:
            span = flask.g.span
            span.record_exception(argv[0])
            if SpanAttributes.HTTP_STATUS_CODE not in span.attributes:
                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 500)
        if flask.g.span.is_recording():
            flask.g.span.end()
        flask.g.span = None

    if hasattr(flask.g, "token") and flask.g.token:
        context.detach(flask.g.token)
        flask.g.token = None


@wrapt.patch_function_wrapper('flask', 'Flask.full_dispatch_request')
def full_dispatch_request_with_instana(
    wrapped: Callable[..., flask.wrappers.Response],
    instance: flask.app.Flask,
    argv: Tuple,
    kwargs: Dict,
) -> flask.wrappers.Response:
    if not hasattr(instance, '_stan_wuz_here'):
        logger.debug("Flask(vanilla): Applying flask before/after instrumentation funcs")
        setattr(instance, "_stan_wuz_here", True)
        instance.before_request(before_request_with_instana)
        instance.after_request(after_request_with_instana)
        instance.teardown_request(teardown_request_with_instana)
    return wrapped(*argv, **kwargs)


logger.debug("Instrumenting flask (without blinker support)")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from typing import Any, Dict, Generator, Optional
import aiohttp
import asyncio

import pytest

from instana.singletons import tracer, agent
from instana.util.ids import hex_id

import tests.apps.flask_app  # noqa: F401
import tests.apps.aiohttp_app  # noqa: F401
from tests.helpers import testenv


class TestAiohttpClient:
    async def fetch(
        self,
        session: aiohttp.client.ClientSession,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        try:
            async with session.get(url, headers=headers, params=params) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        yield
        # teardown
        # Ensure that allow_exit_as_root has the default value"""
        agent.options.allow_exit_as_root = False

    def test_client_get(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_get_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True

        async def test():
            async with aiohttp.ClientSession() as session:
                return await self.fetch(session, testenv["flask_server"] + "/")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        wsgi_span = spans[0]
        aiohttp_span = spans[1]

        # Same traceId
        assert aiohttp_span.t == wsgi_span.t

        # Parent relationships
        assert not aiohttp_span.p
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(wsgi_span.t)}"

    def test_client_get_301(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/301")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        wsgi_span1 = spans[0]
        wsgi_span2 = spans[1]
        aiohttp_span = spans[2]
        test_span = spans[3]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span1.t == traceId
        assert wsgi_span2.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span1.p == aiohttp_span.s
        assert wsgi_span2.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span1.ec
        assert not wsgi_span2.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/301"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span2.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_get_405(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/405")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 405
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/405"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_get_500(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/500")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert aiohttp_span.ec == 1
        assert wsgi_span.ec == 1

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 500
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/500"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["error"] == "INTERNAL SERVER ERROR"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_get_504(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/504")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert aiohttp_span.ec == 1
        assert wsgi_span.ec == 1

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 504
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/504"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["error"] == "GATEWAY TIMEOUT"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_get_with_params_to_scrub(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session, testenv["flask_server"], params={"secret": "yeah"}
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["params"] == "secret=<redacted>"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_client_response_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This"]

        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session, testenv["flask_server"] + "/response_headers"
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/response_headers"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-Capture-This" in aiohttp_span.data["http"]["header"]
        assert aiohttp_span.data["http"]["header"]["X-Capture-This"] == "Ok"

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_client_error(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, "http://doesnotexist:10/")

        response = None
        try:
            response = self.loop.run_until_complete(test())
        except:
            pass

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        aiohttp_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert aiohttp_span.t == test_span.t

        # Parent relationships
        assert aiohttp_span.p == test_span.s

        # Error logging
        assert test_span.ec
        assert aiohttp_span.ec == 1

        assert aiohttp_span.n == "aiohttp-client"
        assert not aiohttp_span.data["http"]["status"]
        assert aiohttp_span.data["http"]["url"] == "http://doesnotexist:10/"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.data["http"]["error"]
        assert len(aiohttp_span.data["http"]["error"])
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert not response

    def test_client_get_tracing_off(self, mocker) -> None:
        mocker.patch(
            "instana.instrumentation.aiohttp.client.tracing_is_off",
            return_value=True,
        )

        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/")

        response = self.loop.run_until_complete(test())
        assert response.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        # Span names are not "aiohttp-client"
        for span in spans:
            assert span.n != "aiohttp-client"

    def test_client_get_provided_tracing_config(self, mocker) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession(trace_configs=[]) as session:
                    return await self.fetch(session, testenv["flask_server"] + "/")

        response = self.loop.run_until_complete(test())
        assert response.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

    def test_client_request_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too"]

        request_headers = {
            "X-Capture-This-Too": "Ok too",
        }

        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session, testenv["flask_server"] + "/", headers=request_headers
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        aiohttp_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aiohttp_span.t == traceId
        assert wsgi_span.t == traceId

        # Parent relationships
        assert aiohttp_span.p == test_span.s
        assert wsgi_span.p == aiohttp_span.s

        # Error logging
        assert not test_span.ec
        assert not aiohttp_span.ec
        assert not wsgi_span.ec

        assert aiohttp_span.n == "aiohttp-client"
        assert aiohttp_span.data["http"]["status"] == 200
        assert aiohttp_span.data["http"]["url"] == testenv["flask_server"] + "/"
        assert aiohttp_span.data["http"]["method"] == "GET"
        assert aiohttp_span.stack
        assert isinstance(aiohttp_span.stack, list)
        assert len(aiohttp_span.stack) > 1

        assert "X-Capture-This-Too" in aiohttp_span.data["http"]["header"]
        assert aiohttp_span.data["http"]["header"]["X-Capture-This-Too"] == "Ok too"

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2024


class Format(object):
    """A namespace for builtin carrier formats.

    These static constants are intended for use in the :meth:`Tracer.inject()`
    and :meth:`Tracer.extract()` methods. E.g.::

        tracer.inject(span.context, Format.BINARY, binary_carrier)

    """

    BINARY = "binary"
    """
    The BINARY format represents SpanContexts in an opaque bytearray carrier.

    For both :meth:`Tracer.inject()` and :meth:`Tracer.extract()` the carrier
    should be a bytearray instance. :meth:`Tracer.inject()` must append to the
    bytearray carrier (rather than replace its contents).
    """

    TEXT_MAP = "text_map"
    """
    The TEXT_MAP format represents :class:`SpanContext`\\ s in a python
    ``dict`` mapping from strings to strings.

    Both the keys and the values have unrestricted character sets (unlike the
    HTTP_HEADERS format).

    NOTE: The TEXT_MAP carrier ``dict`` may contain unrelated data (e.g.,
    arbitrary gRPC metadata). As such, the :class:`Tracer` implementation
    should use a prefix or other convention to distinguish tracer-specific
    key:value pairs.
    """

    HTTP_HEADERS = "http_headers"
    """
    The HTTP_HEADERS format represents :class:`SpanContext`\\ s in a python
    ``dict`` mapping from character-restricted strings to strings.

    Keys and values in the HTTP_HEADERS carrier must be suitable for use as
    HTTP headers (without modification or further escaping). That is, the
    keys have a greatly restricted character set, casing for the keys may not
    be preserved by various intermediaries, and the values should be
    URL-escaped.

    NOTE: The HTTP_HEADERS carrier ``dict`` may contain unrelated data (e.g.,
    arbitrary gRPC metadata). As such, the :class:`Tracer` implementation
    should use a prefix or other convention to distinguish tracer-specific
    key:value pairs.
    """



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import json
import logging
from typing import Generator

import bson
import pymongo
import pytest

from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from tests.helpers import testenv

logger = logging.getLogger(__name__)


class TestPyMongoTracer:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.client = pymongo.MongoClient(
            host=testenv["mongodb_host"],
            port=int(testenv["mongodb_port"]),
            username=testenv["mongodb_user"],
            password=testenv["mongodb_pw"],
        )
        self.client.test.records.delete_many(filter={})
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield
        self.client.close()
        agent.options.allow_exit_as_root = False

    def test_successful_find_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.find_one({"type": "string"})
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "find"

        assert db_span.data["mongo"]["filter"] == '{"type": "string"}'
        assert not db_span.data["mongo"]["json"]

    def test_successful_find_query_as_root_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.client.test.records.find_one({"type": "string"})
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        db_span = spans[0]

        assert not db_span.p
        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "find"

        assert db_span.data["mongo"]["filter"] == '{"type": "string"}'
        assert not db_span.data["mongo"]["json"]

    def test_successful_insert_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.insert_one({"type": "string"})
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "insert"

        assert not db_span.data["mongo"]["filter"]

    def test_successful_update_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.update_one(
                {"type": "string"}, {"$set": {"type": "int"}}
            )
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "update"

        assert not db_span.data["mongo"]["filter"]
        assert db_span.data["mongo"]["json"]

        payload = json.loads(db_span.data["mongo"]["json"])
        assert {
            "q": {"type": "string"},
            "u": {"$set": {"type": "int"}},
            "multi": False,
            "upsert": False,
        } in payload

    def test_successful_delete_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.delete_one(filter={"type": "string"})
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "delete"

        assert not db_span.data["mongo"]["filter"]
        assert db_span.data["mongo"]["json"]

        payload = json.loads(db_span.data["mongo"]["json"])
        assert {"q": {"type": "string"}, "limit": 1} in payload

    def test_successful_aggregate_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.count_documents({"type": "string"})
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert db_span.data["mongo"]["command"] == "aggregate"

        assert not db_span.data["mongo"]["filter"]
        assert db_span.data["mongo"]["json"]

        payload = json.loads(db_span.data["mongo"]["json"])
        assert {"$match": {"type": "string"}} in payload

    @pytest.mark.skipif(
        pymongo.version_tuple >= (4, 0), reason="map reduce is removed in pymongo 4.0"
    )
    def test_successful_map_reduce_query(self) -> None:
        mapper = "function () { this.tags.forEach(function(z) { emit(z, 1); }); }"
        reducer = "function (key, values) { return len(values); }"

        with tracer.start_as_current_span("test"):
            self.client.test.records.map_reduce(
                bson.code.Code(mapper),
                bson.code.Code(reducer),
                "results",
                query={"x": {"$lt": 2}},
            )
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span = spans[0]
        test_span = spans[1]

        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mongo"
        assert (
            db_span.data["mongo"]["service"]
            == f"{testenv['mongodb_host']}:{testenv['mongodb_port']}"
        )
        assert db_span.data["mongo"]["namespace"] == "test.records"
        assert (
            db_span.data["mongo"]["command"].lower() == "mapreduce"
        )  # mapreduce command was renamed to mapReduce in pymongo 3.9.0

        assert db_span.data["mongo"]["filter"] == '{"x": {"$lt": 2}}'
        assert db_span.data["mongo"]["json"]

        payload = json.loads(db_span.data["mongo"]["json"])
        assert payload["map"], {"$code": mapper} == db_span.data["mongo"]["json"]
        assert payload["reduce"], {"$code": reducer} == db_span.data["mongo"]["json"]

    def test_successful_mutiple_queries(self) -> None:
        with tracer.start_as_current_span("test"):
            self.client.test.records.bulk_write(
                [
                    pymongo.InsertOne({"type": "string"}),
                    pymongo.UpdateOne({"type": "string"}, {"$set": {"type": "int"}}),
                    pymongo.DeleteOne({"type": "string"}),
                ]
            )
        current_span = get_current_span()
        assert not current_span.is_recording()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        test_span = spans.pop()

        seen_span_ids = set()
        commands = []
        for span in spans:
            assert test_span.t == span.t
            assert span.p == test_span.s

            # check if all spans got a unique id
            assert span.s not in seen_span_ids

            seen_span_ids.add(span.s)
            commands.append(span.data["mongo"]["command"])

        # ensure spans are ordered the same way as commands
        assert commands == ["insert", "update", "delete"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import importlib
import os
import unittest
import socket

import gevent
from gevent import monkey
from instana import apply_gevent_monkey_patch


class TestGEventAutoTrace(unittest.TestCase):
    def setUp(self):
        # Ensure that the test suite is operational even when Django is installed
        # but not running or configured
        os.environ['DJANGO_SETTINGS_MODULE'] = ''

        self.default_patched_modules = ('socket', 'time', 'select', 'os',
                'threading', 'ssl', 'subprocess', 'signal', 'queue',)

    def tearDown(self):
        if os.environ.get('INSTANA_GEVENT_MONKEY_OPTIONS'):
            os.environ.pop('INSTANA_GEVENT_MONKEY_OPTIONS')

        # Clean up after gevent monkey patches, by restore from the saved dict
        for modname in monkey.saved.keys():
            try:
                mod = __import__(modname)
                importlib.reload(mod)
                for key in monkey.saved[modname].keys():
                    setattr(mod, key, monkey.saved[modname][key])
            except ImportError:
                pass
        monkey.saved = {}


    def test_default_patch_all(self):
        apply_gevent_monkey_patch()
        for module_name in self.default_patched_modules:
            self.assertTrue(monkey.is_module_patched(module_name),
                            f"{module_name} is not patched")

    def test_instana_monkey_options_only_time(self):
        os.environ['INSTANA_GEVENT_MONKEY_OPTIONS'] = (
                'time,no-socket,no-select,no-os,no-select,no-threading,no-os,'
                'no-ssl,no-subprocess,''no-signal,no-queue')
        apply_gevent_monkey_patch()

        self.assertTrue(monkey.is_module_patched('time'), "time module is not patched")
        not_patched_modules = (m for m in self.default_patched_modules if m not in ('time', 'threading'))

        for module_name in not_patched_modules:
            self.assertFalse(monkey.is_module_patched(module_name),
                             f"{module_name} is patched, when it shouldn't be")


    def test_instana_monkey_options_only_socket(self):
        os.environ['INSTANA_GEVENT_MONKEY_OPTIONS'] = (
                '--socket, --no-time, --no-select, --no-os, --no-queue, --no-threading,'
                '--no-os, --no-ssl, no-subprocess, --no-signal, --no-select,')
        apply_gevent_monkey_patch()

        self.assertTrue(monkey.is_module_patched('socket'), "socket module is not patched")
        not_patched_modules = (m for m in self.default_patched_modules if m not in ('socket', 'threading'))

        for module_name in not_patched_modules:
            self.assertFalse(monkey.is_module_patched(module_name),
                             f"{module_name} is patched, when it shouldn't be")




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os

from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles

dir_path = os.path.dirname(os.path.realpath(__file__))


def homepage(request):
    return PlainTextResponse("Hello, world!")


def user(request):
    user_id = request.path_params["user_id"]
    return PlainTextResponse("Hello, user id %s!" % user_id)


def response_headers(request):
    headers = {"X-Capture-This-Too": "this too", "X-Capture-That-Too": "that too"}
    return PlainTextResponse("Stan wuz here with headers!", headers=headers)


async def websocket_endpoint(websocket):
    await websocket.accept()
    await websocket.send_text("Hello, websocket!")
    await websocket.close()


def startup():
    print("Ready to go")


routes = [
    Route("/", homepage),
    Route("/users/{user_id}", user),
    Route("/response_headers", response_headers),
    WebSocketRoute("/ws", websocket_endpoint),
    Mount("/static", StaticFiles(directory=dir_path + "/static")),
]

starlette_server = Starlette(debug=True, routes=routes, on_startup=[startup])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading

from instana.autoprofile.profile import CallSite, Profile
from instana.autoprofile.runtime import RuntimeInfo, min_version
from instana.autoprofile.schedule import schedule
from instana.log import logger

if min_version(3, 4):
    import tracemalloc


class AllocationSampler(object):
    MAX_TRACEBACK_SIZE = 25  # number of frames
    MAX_MEMORY_OVERHEAD = 10 * 1e6  # 10MB
    MAX_PROFILED_ALLOCATIONS = 25

    def __init__(self, profiler: Profile) -> None:
        self.profiler = profiler
        self.ready = False
        self.top = None
        self.top_lock = threading.Lock()
        self.overhead_monitor = None

    def setup(self) -> None:
        if self.profiler.get_option("allocation_sampler_disabled"):
            return

        if not RuntimeInfo.OS_LINUX and not RuntimeInfo.OS_DARWIN:
            logger.debug("Allocation sampler is only supported on Linux and OS X.")
            return

        if not min_version(3, 4):
            logger.debug(
                "Memory allocation profiling is available for Python 3.4 or higher."
            )
            return

        self.ready = True

    def reset(self) -> None:
        self.top = CallSite("", "", 0)

    def start_sampler(self) -> None:
        logger.debug("Activating memory allocation sampler.")

        def start() -> None:
            tracemalloc.start(self.MAX_TRACEBACK_SIZE)

        self.profiler.run_in_main_thread(start)

        def monitor_overhead() -> None:
            if (
                tracemalloc.is_tracing()
                and tracemalloc.get_tracemalloc_memory() > self.MAX_MEMORY_OVERHEAD
            ):
                logger.debug(
                    f"Allocation sampler memory overhead limit exceeded: {tracemalloc.get_tracemalloc_memory()} bytes."
                )
                self.stop_sampler()

        if not self.profiler.get_option("disable_timers"):
            self.overhead_monitor = schedule(0.5, 0.5, monitor_overhead)

    def stop_sampler(self) -> None:
        logger.debug("Deactivating memory allocation sampler.")

        with self.top_lock:
            if self.overhead_monitor:
                self.overhead_monitor.cancel()
                self.overhead_monitor = None

            if tracemalloc.is_tracing():
                snapshot = tracemalloc.take_snapshot()
                logger.debug(
                    f"Allocation sampler memory overhead {tracemalloc.get_tracemalloc_memory()} bytes.",
                )
                tracemalloc.stop()
                self.process_snapshot(snapshot)

    def build_profile(self, duration: int, timespan: int) -> Profile:
        with self.top_lock:
            self.top.normalize(duration)
            self.top.floor()

            profile = Profile(
                Profile.CATEGORY_MEMORY,
                Profile.TYPE_MEMORY_ALLOCATION_RATE,
                Profile.UNIT_BYTE,
                self.top.children.values(),
                duration,
                timespan,
            )

            return profile

    def destroy(self) -> None:
        pass

    def process_snapshot(self, snapshot: tracemalloc.Snapshot) -> None:
        stats = snapshot.statistics("traceback")

        for stat in stats[: self.MAX_PROFILED_ALLOCATIONS]:
            if stat.traceback:
                skip_stack = False
                for frame in stat.traceback:
                    if frame.filename and self.profiler.frame_cache.is_profiler_frame(
                        frame.filename
                    ):
                        skip_stack = True
                        break
                if skip_stack:
                    continue

                current_node = self.top
                for frame in reversed(stat.traceback):
                    if frame.filename == "<unknown>":
                        continue

                    current_node = current_node.find_or_add_child(
                        "", frame.filename, frame.lineno
                    )
                current_node.increment(stat.size, stat.count)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import random
import time
from typing import Generator

import pytest
from cassandra import ConsistencyLevel
from cassandra.cluster import Cluster, ResultSet
from cassandra.query import SimpleStatement

from instana.singletons import agent, tracer
from tests.helpers import get_first_span_by_name, testenv

cluster = Cluster([testenv["cassandra_host"]], load_balancing_policy=None)
session = cluster.connect()

session.execute(
    "CREATE KEYSPACE IF NOT EXISTS instana_tests WITH replication = {'class':'SimpleStrategy', 'replication_factor':1};"
)
session.set_keyspace("instana_tests")
session.execute(
    "CREATE TABLE IF NOT EXISTS users("
    "id int PRIMARY KEY,"
    "name text,"
    "age text,"
    "email varint,"
    "phone varint"
    ");"
)


class TestCassandra:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield
        agent.options.allow_exit_as_root = False

    def test_untraced_execute(self) -> None:
        res = session.execute("SELECT name, age, email FROM users")

        assert isinstance(res, ResultSet)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 0

    def test_untraced_execute_error(self) -> None:
        res = None
        try:
            res = session.execute("Not a valid query")
        except Exception:
            pass

        assert not res

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 0

    def test_execute(self) -> None:
        res = None
        with tracer.start_as_current_span("test"):
            res = session.execute("SELECT name, age, email FROM users")

        assert isinstance(res, ResultSet)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        # Same traceId and parent relationship
        assert cspan.t == test_span.t
        assert cspan.p == test_span.s

        assert cspan.stack
        assert not cspan.ec

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert cspan.data["cassandra"]["query"] == "SELECT name, age, email FROM users"
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert not cspan.data["cassandra"]["achievedConsistency"]
        assert cspan.data["cassandra"]["triedHosts"]
        assert not cspan.data["cassandra"]["error"]

    def test_execute_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        res = session.execute("SELECT name, age, email FROM users")

        assert isinstance(res, ResultSet)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        assert not cspan.p

        assert cspan.stack
        assert not cspan.ec

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert cspan.data["cassandra"]["query"] == "SELECT name, age, email FROM users"
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert not cspan.data["cassandra"]["achievedConsistency"]
        assert cspan.data["cassandra"]["triedHosts"]
        assert not cspan.data["cassandra"]["error"]

    def test_execute_async(self) -> None:
        res = None
        with tracer.start_as_current_span("test"):
            res = session.execute_async("SELECT name, age, email FROM users").result()

        assert isinstance(res, ResultSet)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        # Same traceId and parent relationship
        assert cspan.t == test_span.t
        assert cspan.p == test_span.s

        assert cspan.stack
        assert not cspan.ec

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert cspan.data["cassandra"]["query"] == "SELECT name, age, email FROM users"
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert not cspan.data["cassandra"]["achievedConsistency"]
        assert cspan.data["cassandra"]["triedHosts"]
        assert not cspan.data["cassandra"]["error"]

    def test_simple_statement(self) -> None:
        res = None
        with tracer.start_as_current_span("test"):
            query = SimpleStatement(
                "SELECT name, age, email FROM users", is_idempotent=True
            )
            res = session.execute(query)

        assert isinstance(res, ResultSet)

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        # Same traceId and parent relationship
        assert cspan.t == test_span.t
        assert cspan.p == test_span.s

        assert cspan.stack
        assert not cspan.ec

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert cspan.data["cassandra"]["query"] == "SELECT name, age, email FROM users"
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert not cspan.data["cassandra"]["achievedConsistency"]
        assert cspan.data["cassandra"]["triedHosts"]
        assert not cspan.data["cassandra"]["error"]

    def test_execute_error(self) -> None:
        res = None

        try:
            with tracer.start_as_current_span("test"):
                res = session.execute("Not a real query")
        except Exception:
            pass

        assert not res

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        # Same traceId and parent relationship
        assert cspan.t == test_span.t
        assert cspan.p == test_span.s

        assert cspan.stack
        assert cspan.ec == 1

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert cspan.data["cassandra"]["query"] == "Not a real query"
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert not cspan.data["cassandra"]["achievedConsistency"]
        assert cspan.data["cassandra"]["triedHosts"]
        assert cspan.data["cassandra"]["error"] == "Syntax error in CQL query"

    def test_prepared_statement(self) -> None:
        prepared = None

        with tracer.start_as_current_span("test"):
            prepared = session.prepare(
                "INSERT INTO users (id, name, age) VALUES (?, ?, ?)"
            )
            prepared.consistency_level = ConsistencyLevel.QUORUM
            session.execute(prepared, (random.randint(0, 1000000), "joe", "17"))

        assert prepared

        time.sleep(0.5)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        test_span = get_first_span_by_name(spans, "sdk")
        assert test_span
        assert test_span.data["sdk"]["name"] == "test"

        cspan = get_first_span_by_name(spans, "cassandra")
        assert cspan

        # Same traceId and parent relationship
        assert test_span.t == cspan.t
        assert cspan.p == test_span.s

        assert cspan.stack
        assert not cspan.ec

        assert cspan.data["cassandra"]["cluster"] == "Test Cluster"
        assert (
            cspan.data["cassandra"]["query"]
            == "INSERT INTO users (id, name, age) VALUES (?, ?, ?)"
        )
        assert cspan.data["cassandra"]["keyspace"] == "instana_tests"
        assert cspan.data["cassandra"]["achievedConsistency"] == "QUORUM"
        assert cspan.data["cassandra"]["triedHosts"]
        assert not cspan.data["cassandra"]["error"]



# coding: utf-8
# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

try:
    import types
    from typing import (
        TYPE_CHECKING,
        Any,
        Callable,
        Dict,
        Iterator,
        Optional,
        Tuple,
        Union,
    )

    import pika
    import wrapt

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.singletons import tracer
    from instana.util.traceutils import get_tracer_tuple, tracing_is_off

    if TYPE_CHECKING:
        import pika.adapters.blocking_connection
        import pika.channel
        import pika.connection

        from instana.span.span import InstanaSpan

    def _extract_broker_attributes(
        span: "InstanaSpan", conn: pika.connection.Connection
    ) -> None:
        span.set_attribute("address", f"{conn.params.host}:{conn.params.port}")

    def _extract_publisher_attributes(
        span: "InstanaSpan",
        conn: pika.connection.Connection,
        exchange: str,
        routing_key: str,
    ) -> None:
        _extract_broker_attributes(span, conn)

        span.set_attribute("sort", "publish")
        span.set_attribute("key", routing_key)
        span.set_attribute("exchange", exchange)

    def _extract_consumer_tags(
        span: "InstanaSpan", conn: pika.connection.Connection, queue: str
    ) -> None:
        _extract_broker_attributes(span, conn)

        span.set_attribute("sort", "consume")
        span.set_attribute("queue", queue)

    @wrapt.patch_function_wrapper("pika.channel", "Channel.basic_publish")
    def basic_publish_with_instana(
        wrapped: Callable[..., pika.channel.Channel.basic_publish],
        instance: pika.channel.Channel,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        def _bind_args(
            exchange: str,
            routing_key: str,
            body: str,
            properties: Optional[object] = None,
            *args: object,
            **kwargs: object,
        ) -> Tuple[object, ...]:
            return (exchange, routing_key, body, properties, args, kwargs)

        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        (exchange, routing_key, body, properties, args, kwargs) = _bind_args(
            *args, **kwargs
        )

        with tracer.start_as_current_span(
            "rabbitmq", span_context=parent_context
        ) as span:
            try:
                _extract_publisher_attributes(
                    span,
                    conn=instance.connection,
                    routing_key=routing_key,
                    exchange=exchange,
                )
            except Exception:
                logger.debug("pika publish_with_instana error: ", exc_info=True)

            # context propagation
            properties = properties or pika.BasicProperties()
            properties.headers = properties.headers or {}

            tracer.inject(
                span.context,
                Format.HTTP_HEADERS,
                properties.headers,
                disable_w3c_trace_context=True,
            )
            args = (exchange, routing_key, body, properties) + args

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    def basic_get_with_instana(
        wrapped: Callable[
            ...,
            Union[pika.channel.Channel.basic_get, pika.channel.Channel.basic_consume],
        ],
        instance: pika.channel.Channel,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        def _bind_args(*args: object, **kwargs: object) -> Tuple[object, ...]:
            args = list(args)
            queue = kwargs.pop("queue", None) or args.pop(0)
            callback = (
                kwargs.pop("callback", None)
                or kwargs.pop("on_message_callback", None)
                or args.pop(0)
            )
            return (queue, callback, tuple(args), kwargs)

        queue, callback, args, kwargs = _bind_args(*args, **kwargs)

        def _cb_wrapper(
            channel: pika.channel.Channel,
            method: pika.spec.Basic,
            properties: pika.BasicProperties,
            body: str,
        ) -> None:
            parent_context = tracer.extract(
                Format.HTTP_HEADERS, properties.headers, disable_w3c_trace_context=True
            )

            with tracer.start_as_current_span(
                "rabbitmq", span_context=parent_context
            ) as span:
                try:
                    _extract_consumer_tags(span, conn=instance.connection, queue=queue)
                except Exception:
                    logger.debug("pika basic_get_with_instana error: ", exc_info=True)

                try:
                    callback(channel, method, properties, body)
                except Exception as exc:
                    span.record_exception(exc)

        args = (queue, _cb_wrapper) + args
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper(
        "pika.adapters.blocking_connection", "BlockingChannel.basic_consume"
    )
    def basic_consume_with_instana(
        wrapped: Callable[
            ..., pika.adapters.blocking_connection.BlockingChannel.basic_consume
        ],
        instance: pika.adapters.blocking_connection.BlockingChannel,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        def _bind_args(
            queue: str,
            on_message_callback: object,
            *args: object,
            **kwargs: object,
        ) -> Tuple[object, ...]:
            return (queue, on_message_callback, args, kwargs)

        queue, on_message_callback, args, kwargs = _bind_args(*args, **kwargs)

        def _cb_wrapper(
            channel: pika.channel.Channel,
            method: pika.spec.Basic,
            properties: pika.BasicProperties,
            body: str,
        ) -> None:
            parent_context = tracer.extract(
                Format.HTTP_HEADERS, properties.headers, disable_w3c_trace_context=True
            )

            with tracer.start_as_current_span(
                "rabbitmq", span_context=parent_context
            ) as span:
                try:
                    _extract_consumer_tags(
                        span, conn=instance.connection._impl, queue=queue
                    )
                except Exception:
                    logger.debug(
                        "pika basic_consume_with_instana error:", exc_info=True
                    )

                try:
                    on_message_callback(channel, method, properties, body)
                except Exception as exc:
                    span.record_exception(exc)

        args = (queue, _cb_wrapper) + args
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper(
        "pika.adapters.blocking_connection", "BlockingChannel.consume"
    )
    def consume_with_instana(
        wrapped: Callable[..., pika.adapters.blocking_connection.BlockingChannel],
        instance: pika.adapters.blocking_connection.BlockingChannel,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        def _bind_args(
            queue: str, *args: object, **kwargs: object
        ) -> Tuple[object, ...]:
            return (queue, args, kwargs)

        (queue, args, kwargs) = _bind_args(*args, **kwargs)

        def _consume(gen: Iterator[object]) -> object:
            for yielded in gen:
                # Bypass the delivery created due to inactivity timeout
                if not yielded or not any(yielded):
                    yield yielded
                    continue

                (method_frame, properties, body) = yielded

                parent_context = tracer.extract(
                    Format.HTTP_HEADERS,
                    properties.headers,
                    disable_w3c_trace_context=True,
                )
                with tracer.start_as_current_span(
                    "rabbitmq", span_context=parent_context
                ) as span:
                    try:
                        _extract_consumer_tags(
                            span, conn=instance.connection._impl, queue=queue
                        )
                    except Exception:
                        logger.debug("consume_with_instana: ", exc_info=True)

                    try:
                        yield yielded
                    except Exception as exc:
                        span.record_exception(exc)

        args = (queue,) + args
        res = wrapped(*args, **kwargs)

        if isinstance(res, types.GeneratorType):
            return _consume(res)
        else:
            return res

    @wrapt.patch_function_wrapper(
        "pika.adapters.blocking_connection", "BlockingChannel.__init__"
    )
    def _BlockingChannel___init__(
        wrapped: Callable[
            ..., pika.adapters.blocking_connection.BlockingChannel.__init__
        ],
        instance: pika.adapters.blocking_connection.BlockingChannel,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        ret = wrapped(*args, **kwargs)
        impl = getattr(instance, "_impl", None)

        if impl and hasattr(impl.basic_consume, "__wrapped__"):
            impl.basic_consume = impl.basic_consume.__wrapped__

        return ret

    wrapt.wrap_function_wrapper(
        "pika.channel", "Channel.basic_get", basic_get_with_instana
    )
    wrapt.wrap_function_wrapper(
        "pika.channel", "Channel.basic_consume", basic_get_with_instana
    )

    logger.debug("Instrumenting pika")
except ImportError:
    pass
